@@ -3,7 +3,13 @@
 {-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE TupleSections   #-}
 
--- | This module contains functions and data types to parse CLI inputs.
+{- |
+Copyright: (c) 2017-2019 Kowainik
+SPDX-License-Identifier: MPL-2.0
+Maintainer: Kowainik <xrom.xkov@gmail.com>
+
+This module contains functions and data types to parse CLI inputs.
+-}
 
 module Summoner.CLI
        ( -- * CLI data types
@@ -16,36 +22,43 @@ module Summoner.CLI
        , summonCli
 
          -- * Runners
+       , runConfig
        , runScript
 
          -- * Common helper functions
        , getFinalConfig
+       , getCustomLicenseText
        ) where
 
+import Colourista (blue, bold, errorMessage, formatWith, infoMessage, successMessage,
+                   warningMessage)
 import Data.Version (Version, showVersion)
-import Development.GitRev (gitCommitDate, gitDirty, gitHash)
+import Development.GitRev (gitCommitDate, gitHash)
 import NeatInterpolation (text)
-import Options.Applicative (Parser, ParserInfo, argument, command, execParser, flag, fullDesc, help,
-                            helper, info, infoFooter, infoHeader, infoOption, long, maybeReader,
-                            metavar, option, optional, progDesc, short, strArgument, strOption,
-                            subparser, switch, value)
+import Options.Applicative (Parser, ParserInfo, ParserPrefs, argument, command, customExecParser,
+                            flag, fullDesc, help, helpLongEquals, helper, info, infoFooter,
+                            infoHeader, infoOption, long, maybeReader, metavar, option, prefs,
+                            progDesc, short, showDefault, showHelpOnEmpty, strArgument, strOption,
+                            subparser, subparserInline, switch, value)
 import Options.Applicative.Help.Chunk (stringChunk)
-import Shellmet (($?), ($|))
+import Relude.Extra.Enum (universe)
 import System.Directory (doesFileExist)
 import System.Info (os)
+import Validation (Validation (..))
 
-import Summoner.Ansi (blueCode, boldCode, errorMessage, infoMessage, redCode, resetCode,
-                      successMessage, warningMessage)
 import Summoner.Config (Config, ConfigP (..), PartialConfig, defaultConfig, finalise,
-                        loadFileConfig)
+                        guessConfigFromGit, loadFileConfig)
+import Summoner.CustomPrelude (CustomPrelude (..))
 import Summoner.Decision (Decision (..))
-import Summoner.Default (defaultConfigFile, defaultGHC)
-import Summoner.GhcVer (GhcVer, parseGhcVer, showGhcVer)
-import Summoner.License (License (..), LicenseName (..), fetchLicense, parseLicenseName,
+import Summoner.Default (currentYear, defaultConfigFile, defaultConfigFileContent, defaultGHC)
+import Summoner.GhcVer (GhcVer, ghcTable, parseGhcVer)
+import Summoner.License (License (..), LicenseName (..), fetchLicenseCustom, parseLicenseName,
                          showLicenseWithDesc)
+import Summoner.Mode (ConnectMode (..), Interactivity (..), isOffline)
 import Summoner.Project (generateProject)
-import Summoner.Settings (CustomPrelude (..), Tool, parseTool)
+import Summoner.Settings (Tool, parseTool)
 import Summoner.Template.Script (scriptFile)
+import Summoner.Text (endLine)
 
 import qualified Data.Text as T
 import qualified Paths_summoner as Meta (version)
@@ -54,7 +67,15 @@ import qualified Paths_summoner as Meta (version)
 -- | Main function that parses @CLI@ commands and runs them using given
 -- 'Command' handler.
 summon :: Version -> (Command -> IO ()) -> IO ()
-summon version performCommand = execParser (cliParser version) >>= performCommand
+summon version performCommand =
+    customExecParser summonerParserPrefs (cliParser version) >>= performCommand
+  where
+    -- To turn on some special options.
+    summonerParserPrefs :: ParserPrefs
+    summonerParserPrefs = prefs
+        $ helpLongEquals
+       <> showHelpOnEmpty
+       <> subparserInline
 
 -- | Runs @summoner@ in CLI mode.
 summonCli :: IO ()
@@ -66,6 +87,35 @@ runCliCommand = \case
     New opts -> runNew opts
     Script opts -> runScript opts
     ShowInfo opts -> runShow opts
+    Config opts -> runConfig opts
+
+
+{- | Runs @config@ command
+
+@
+Usage: summon config [-f|--file=FILENAME]
+  Create a default TOML configuration file for summoner
+
+Available options:
+  -h,--help                Show this help text
+  -f,--file=FILENAME       Path to the toml file with configurations. If not
+                           specified '~/.summoner.toml' will be used by default
+@
+-}
+runConfig :: ConfigOpts -> IO ()
+runConfig ConfigOpts{..} = do
+    configFile <- whenNothing configOptsName defaultConfigFile
+    let configFileTxt = toText configFile
+    alreadyExist <- doesFileExist configFile
+    if alreadyExist
+    then do
+        warningMessage $ "File '" <> configFileTxt <> "' already exits."
+        infoMessage "Use 'summon config --file <path>' to specify another path."
+        exitFailure
+    else do
+        writeFileText configFile defaultConfigFileContent
+        infoMessage $ "Created default configuration file: " <> configFileTxt
+        infoMessage "Open the file using the editor of your choice."
 
 {- | Runs @show@ command.
 
@@ -84,7 +134,7 @@ Available commands:
 runShow :: ShowOpts -> IO ()
 runShow = \case
     -- show list of all available GHC versions
-    GhcList -> showBulletList @GhcVer showGhcVer (reverse universe)
+    GhcList -> showBulletList id ghcTable
     -- show a list of all available licenses
     LicenseList Nothing -> showBulletList @LicenseName showLicenseWithDesc universe
     -- show a specific license
@@ -96,12 +146,22 @@ runShow = \case
                 showBulletList @LicenseName show universe
                 -- get and show a license`s text
             Just licenseName -> do
-                fetchedLicense <- fetchLicense licenseName
-                putTextLn $ unLicense fetchedLicense
+                licenseCustomText <- getCustomLicenseText licenseName
+                putTextLn $ unLicense licenseCustomText
   where
     showBulletList :: (a -> Text) -> [a] -> IO ()
     showBulletList showT = mapM_ (infoMessage . T.append "➤ " . showT)
 
+-- | Get the customized License text for @summon show license NAME@ command.
+getCustomLicenseText :: LicenseName -> IO License
+getCustomLicenseText licenseName = do
+    year <- currentYear
+    guessConfig <- guessConfigFromGit
+    fetchLicenseCustom
+        licenseName
+        (fromMaybe "YOUR NAME" $ getLast $ cFullName guessConfig)
+        year
+
 {- | Runs @script@ command.
 
 @
@@ -147,7 +207,7 @@ runNew newOpts@NewOpts{..} = do
     -- get the final config
     finalConfig <- getFinalConfig newOpts
     -- Generate the project.
-    generateProject newOptsOffline newOptsProjectName finalConfig
+    generateProject newOptsInteractivity newOptsConnectMode newOptsProjectName finalConfig
 
 -- | By the given 'NewOpts' return the final configurations.
 getFinalConfig :: NewOpts -> IO Config
@@ -156,7 +216,7 @@ getFinalConfig NewOpts{..} = do
     fileConfig <- readFileConfig newOptsIgnoreFile newOptsConfigFile
 
     -- guess config from the global ~/.gitconfig file
-    gitConfig <- guessFromGit
+    gitConfig <- guessConfigFromGit
 
     -- union all possible configs
     let unionConfig = defaultConfig <> gitConfig <> fileConfig <> newOptsCliConfig
@@ -165,18 +225,6 @@ getFinalConfig NewOpts{..} = do
     case finalise unionConfig of
         Success c    -> pure c
         Failure msgs -> for_ msgs errorMessage >> exitFailure
-  where
-    guessFromGit :: IO PartialConfig
-    guessFromGit = do
-       gitOwner <- (Just <$> "git" $| ["config", "user.login"]) $? pure Nothing
-       gitName  <- (Just <$> "git" $| ["config", "user.name"])  $? pure Nothing
-       gitEmail <- (Just <$> "git" $| ["config", "user.email"]) $? pure Nothing
-       pure $ defaultConfig
-           { cOwner = Last gitOwner
-           , cFullName = Last gitName
-           , cEmail = Last gitEmail
-           }
-
 
 -- | Reads and parses the given config file. If no file is provided the default
 -- configuration returned.
@@ -211,17 +259,20 @@ data Command
     | Script ScriptOpts
     -- | @show@ command shows supported licenses or GHC versions
     | ShowInfo ShowOpts
+    -- | @config@ command creates the TOML configuration file
+    | Config ConfigOpts
 
--- | Options parsed with @new@ command
+-- | Options parsed with the @new@ command
 data NewOpts = NewOpts
-    { newOptsProjectName :: Text           -- ^ project name
-    , newOptsIgnoreFile  :: Bool           -- ^ ignore all config files if 'True'
-    , newOptsOffline     :: Bool           -- ^ Offline mode
-    , newOptsConfigFile  :: Maybe FilePath -- ^ file with custom configuration
-    , newOptsCliConfig   :: PartialConfig  -- ^ config gathered during CLI
+    { newOptsProjectName   :: !Text             -- ^ Project name
+    , newOptsIgnoreFile    :: !Bool             -- ^ Ignore all config files if 'True'
+    , newOptsConnectMode   :: !ConnectMode      -- ^ 'Online'/'Offline' mode
+    , newOptsInteractivity :: !Interactivity    -- ^ Interactive or non-interactive mode is on?
+    , newOptsConfigFile    :: !(Maybe FilePath) -- ^ File with custom configuration
+    , newOptsCliConfig     :: !PartialConfig    -- ^ Config gathered via command-line
     }
 
--- | Options parsed with @script@ command
+-- | Options parsed with the @script@ command
 data ScriptOpts = ScriptOpts
     { scriptOptsTool :: !Tool      -- ^ Build tool: `cabal` or `stack`
     , scriptOptsName :: !FilePath  -- ^ File path to the script
@@ -233,6 +284,11 @@ data ShowOpts
     = GhcList
     | LicenseList (Maybe String)
 
+-- | Options parsed with the @config@ command
+newtype ConfigOpts = ConfigOpts
+    { configOptsName :: Maybe FilePath
+    }
+
 ----------------------------------------------------------------------------
 -- Parsers
 ----------------------------------------------------------------------------
@@ -252,21 +308,30 @@ versionP version = infoOption (summonerVersion version)
    <> help "Show summoner's version"
 
 summonerVersion :: Version -> String
-summonerVersion version = toString
-    $ intercalate "\n"
-    $ [sVersion, sHash, sDate] ++ [sDirty | $(gitDirty)]
+summonerVersion version = intercalate "\n" [sVersion, sHash, sDate]
   where
-    sVersion = blueCode <> boldCode <> "Summoner " <> "v" <>  showVersion version <> resetCode
-    sHash = " ➤ " <> blueCode <> boldCode <> "Git revision: " <> resetCode <> $(gitHash)
-    sDate = " ➤ " <> blueCode <> boldCode <> "Commit date:  " <> resetCode <> $(gitCommitDate)
-    sDirty = redCode <> "There are non-committed files." <> resetCode
+    sVersion, sHash, sDate :: String
+    sVersion = formatWith [blue, bold] $ "Summoner " <> "v" <>  showVersion version
+    sHash = " ➤ " <> formatWith [blue, bold] "Git revision: " <> $(gitHash)
+    sDate = " ➤ " <> formatWith [blue, bold] "Commit date:  " <> $(gitCommitDate)
 
--- All possible commands.
+-- | All possible commands.
 summonerP :: Parser Command
 summonerP = subparser
     $ command "new" (info (helper <*> newP) $ progDesc "Create a new Haskell project")
    <> command "script" (info (helper <*> scriptP) $ progDesc "Create a new Haskell script")
    <> command "show" (info (helper <*> showP) $ progDesc "Show supported licenses or ghc versions")
+   <> command "config" (info (helper <*> configP) $ progDesc "Create a default TOML configuration file for summoner")
+
+----------------------------------------------------------------------------
+-- @config@ command parsers
+----------------------------------------------------------------------------
+
+-- | Parses options of the @config@ command.
+configP :: Parser Command
+configP = do
+    configOptsName <- optional configFileP
+    pure $ Config ConfigOpts{..}
 
 ----------------------------------------------------------------------------
 -- @show@ command parsers
@@ -313,6 +378,7 @@ ghcVerP = option
     (  long "ghc"
     <> short 'g'
     <> value defaultGHC
+    <> showDefault
     <> metavar "GHC_VERSION"
     <> help "Version of the compiler to be used for script"
     )
@@ -324,11 +390,12 @@ ghcVerP = option
 -- | Parses options of the @new@ command.
 newP :: Parser Command
 newP = do
-    newOptsProjectName <- strArgument (metavar "PROJECT_NAME")
-    newOptsIgnoreFile  <- ignoreFileP
-    noUpload           <- noUploadP
-    newOptsOffline     <- offlineP
-    newOptsConfigFile  <- optional fileP
+    newOptsProjectName   <- strArgument (metavar "PROJECT_NAME")
+    newOptsIgnoreFile    <- ignoreFileP
+    noUpload             <- noUploadP
+    newOptsConnectMode   <- connectModeP
+    newOptsInteractivity <- interactivityP
+    newOptsConfigFile    <- optional configFileP
     cabal <- cabalP
     stack <- stackP
     preludePack <- optional preludePackP
@@ -341,79 +408,96 @@ newP = do
             { cPrelude = Last $ CustomPrelude <$> preludePack <*> preludeMod
             , cCabal = cabal
             , cStack = stack
-            , cNoUpload = Any $ noUpload || newOptsOffline
+            , cNoUpload = Any $ noUpload || isOffline newOptsConnectMode
             }
         , ..
         }
 
 targetsP ::  Decision -> Parser PartialConfig
 targetsP d = do
-    cGitHub  <- githubP    d
-    cTravis  <- travisP    d
-    cAppVey  <- appVeyorP  d
-    cPrivate <- privateP   d
-    cLib     <- libraryP   d
-    cExe     <- execP      d
-    cTest    <- testP      d
-    cBench   <- benchmarkP d
+    cGitHub    <- githubP    d
+    cGhActions <- ghActionsP d
+    cTravis    <- travisP    d
+    cAppVey    <- appVeyorP  d
+    cPrivate   <- privateP   d
+    cLib       <- libraryP   d
+    cExe       <- execP      d
+    cTest      <- testP      d
+    cBench     <- benchmarkP d
     pure mempty
-        { cGitHub = cGitHub
-        , cTravis = cTravis
-        , cAppVey = cAppVey
-        , cPrivate= cPrivate
-        , cLib    = cLib
-        , cExe    = cExe
-        , cTest   = cTest
-        , cBench  = cBench
+        { cGitHub    = cGitHub
+        , cGhActions = cGhActions
+        , cTravis    = cTravis
+        , cAppVey    = cAppVey
+        , cPrivate   = cPrivate
+        , cLib       = cLib
+        , cExe       = cExe
+        , cTest      = cTest
+        , cBench     = cBench
         }
 
 githubP :: Decision -> Parser Decision
-githubP d = flag Idk d
-          $ long "github"
-         <> short 'g'
-         <> help "GitHub integration"
+githubP d = flag Idk d $ mconcat
+    [ long "github"
+    , short 'g'
+    , help "GitHub integration"
+    ]
+
+ghActionsP :: Decision -> Parser Decision
+ghActionsP d = flag Idk d $ mconcat
+    [ long "actions"
+    , short 'a'
+    , help "GitHub Actions CI"
+    ]
 
 travisP :: Decision -> Parser Decision
-travisP d = flag Idk d
-          $ long "travis"
-         <> short 'c'
-         <> help "Travis CI integration"
+travisP d = flag Idk d $ mconcat
+    [ long "travis"
+    , short 'c'
+    , help "Travis CI integration"
+    ]
 
 appVeyorP :: Decision -> Parser Decision
-appVeyorP d = flag Idk d
-            $ long "app-veyor"
-           <> short 'w'
-           <> help "AppVeyor CI integration"
+appVeyorP d = flag Idk d $ mconcat
+    [ long "app-veyor"
+    , short 'w'
+    , help "AppVeyor CI integration"
+    ]
 
 privateP :: Decision -> Parser Decision
-privateP d = flag Idk d
-           $ long "private"
-          <> short 'p'
-          <> help "Private repository"
+privateP d = flag Idk d $ mconcat
+    [ long "private"
+    , short 'p'
+    , help "Private repository"
+    ]
 
 libraryP :: Decision -> Parser Decision
-libraryP d = flag Idk d
-           $ long "library"
-          <> short 'l'
-          <> help "Library folder"
+libraryP d = flag Idk d $ mconcat
+    [ long "library"
+    , short 'l'
+    , help "Library folder"
+    ]
 
 execP :: Decision -> Parser Decision
-execP d = flag Idk d
-        $ long "exec"
-       <> short 'e'
-       <> help "Executable target"
+execP d = flag Idk d $ mconcat
+    [ long "exec"
+    , short 'e'
+    , help "Executable target"
+    ]
 
 testP :: Decision -> Parser Decision
-testP d = flag Idk d
-        $ long "test"
-       <> short 't'
-       <> help "Test target"
+testP d = flag Idk d $ mconcat
+    [ long "test"
+    , short 't'
+    , help "Test target"
+    ]
 
 benchmarkP :: Decision -> Parser Decision
-benchmarkP d = flag Idk d
-             $ long "benchmark"
-            <> short 'b'
-            <> help "Benchmarks"
+benchmarkP d = flag Idk d $ mconcat
+    [ long "benchmark"
+    , short 'b'
+    , help "Benchmarks"
+    ]
 
 withP :: Parser PartialConfig
 withP = subparser $ mconcat
@@ -428,44 +512,63 @@ withoutP = subparser $ mconcat
     ]
 
 ignoreFileP :: Parser Bool
-ignoreFileP = switch $ long "ignore-config" <> help "Ignore configuration file"
+ignoreFileP = switch $ mconcat
+    [ long "ignore-config"
+    , help "Ignore configuration file"
+    ]
 
 noUploadP :: Parser Bool
-noUploadP = switch $ long "no-upload" <> help "Do not upload to GitHub. Special case of the '--offline' flag."
+noUploadP = switch $ mconcat
+    [ long "no-upload"
+    , help "Do not upload to GitHub. Special case of the '--offline' flag."
+    ]
 
-offlineP :: Parser Bool
-offlineP = switch
-    $ long "offline"
-   <> help "Offline mode: create project with 'All Rights Reserved' license and without uploading to GitHub."
+connectModeP :: Parser ConnectMode
+connectModeP = flag Online Offline $ mconcat
+    [ long "offline"
+    , help "Offline mode: create a project with 'All Rights Reserved' license and without uploading to GitHub."
+    ]
 
-fileP :: Parser FilePath
-fileP = strOption
-    $ long "file"
-   <> short 'f'
-   <> metavar "FILENAME"
-   <> help "Path to the toml file with configurations. If not specified '~/.summoner.toml' will be used if present"
+interactivityP :: Parser Interactivity
+interactivityP = flag Interactive NonInteractive $ mconcat
+    [ long "non-interactive"
+    , short 'n'
+    , help "Non-interactive mode: create a project without interactive questions."
+    ]
+
+configFileP :: Parser FilePath
+configFileP = strOption $ mconcat
+    [ long "file"
+    , short 'f'
+    , metavar "FILENAME"
+    , help "Path to the toml file with configurations. If not specified '~/.summoner.toml' will be used by default"
+    ]
 
 preludePackP :: Parser Text
-preludePackP = strOption
-    $ long "prelude-package"
-   <> metavar "PACKAGE_NAME"
-   <> help "Name for the package of the custom prelude to use in the project"
+preludePackP = strOption $ mconcat
+    [ long "prelude-package"
+    , metavar "PACKAGE_NAME"
+    , help "Name for the package of the custom prelude to use in the project"
+    ]
 
 preludeModP :: Parser Text
-preludeModP = strOption
-    $ long "prelude-module"
-   <> metavar "MODULE_NAME"
-   <> help "Name for the module of the custom prelude to use in the project"
+preludeModP = strOption $ mconcat
+    [ long "prelude-module"
+    , metavar "MODULE_NAME"
+    , help "Name for the module of the custom prelude to use in the project"
+    ]
 
 cabalP :: Parser Decision
-cabalP = flag Idk Yes
-       $ long "cabal"
-      <> help "Cabal support for the project"
+cabalP = flag Idk Yes $ mconcat
+    [ long "cabal"
+    , help "Cabal support for the project"
+    ]
 
 stackP :: Parser Decision
-stackP = flag Idk Yes
-       $ long "stack"
-      <> help "Stack support for the project"
+stackP = flag Idk Yes $ mconcat
+    [ long "stack"
+    , help "Stack support for the project"
+    ]
 
 ----------------------------------------------------------------------------
 -- Beauty util
